import sys
import os
import pathlib

from mlflow.tracking import MlflowClient
import mlflow.pyfunc
import mlflow


from apscheduler.schedulers.background import BackgroundScheduler
import datetime


# FastAPI app
from fastapi import FastAPI, Request, HTTPException, Depends
import fastapi
from fastapi.openapi.utils import get_openapi
from fastapi.security import HTTPBearer
import re
from starlette.routing import RedirectResponse
from starlette.routing import Route as starlette_Route
import urllib
from typing import Dict, List, Any

# internal requests for artifact
import requests

# logging
from fastapi.logger import logger
import logging
import traceback


from .config import p as cfg
from .basehandler import BaseHandler, load
from .basehandler import load as load_BaseHandler

__version__ = "0.1.8"


class Server:

    def __init__(self, config=None):
        if config is not None:
            self.config = config
        else:
            self.config = cfg.parse_known_args()[0]

        # create the dictionary with all the models
        self.model_dict = {}
        # create a dictionary with all errors
        self.error_dict = {}

        # init logger
        gunicorn_logger = logging.getLogger('gunicorn.error')
        logger.handlers = gunicorn_logger.handlers
        logger.setLevel(gunicorn_logger.level)
        self.logger = logger
        logger.info(f"Start server version {__version__}")

        # connect to mlflow
        if "databricks" in self.config.mlflow:
            os.environ['DATABRICKS_HOST'] = self.config.mlflow
            os.environ['DATABRICKS_TOKEN'] = self.config.mlflow_token
        if self.config.mlflow_noverify:
            os.environ["MLFLOW_TRACKING_INSECURE_TLS"] = "true"
        mlflow.set_tracking_uri(self.config.mlflow)
        if self.config.mlflow_token:
            os.environ["MLFLOW_TRACKING_TOKEN"] = self.config.mlflow_token
        self.client = MlflowClient()

        # register at eureka
        if self.config.eureka_server:
            try:
                import py_eureka_client.eureka_client as eureka_client
                eureka_client.init(eureka_server=self.config.eureka_server,
                                app_name=self.config.app_name,
                                instance_port=self.config.port,
                                instance_host=self.config.host_name
                                )
            except Exception as ex:
                self.error_dict["eureka"] = {
                    "message": str(ex),
                }

        # init schedulerr
        self.scheduler = None
        self.init_scheduler()

        # init FastAPI app
        tags_metadata = [
            {
                "name": "Metadata",
                "description": "Get meta information of the models.",
            },
            {
                "name": "Models",
                "description": "Predict with the models.",
            }
        ]
        self.app = FastAPI(
            redoc_url=None, tags_metadata=tags_metadata, root_path=self.config.basepath)
        self.security = HTTPBearer()

        def custom_openapi():
            if self.app.openapi_schema:
                return self.app.openapi_schema
            openapi_schema = get_openapi(
                title=self.config.title,
                version=__version__,
                description=self.config.description,
                routes=self.app.routes,
            )
            self.app.openapi_schema = openapi_schema
            return self.app.openapi_schema
        self.app.openapi = custom_openapi

        # move all routes to the basepath
        if self.config.basepath != "":
            basepath_re = self.config.basepath.replace("/", "\\/")+"\\/"
            for r in self.app.routes:
                if not isinstance(r, fastapi.routing.APIRoute):
                    r.path_regex = re.compile(r.path_regex.pattern.replace(
                        '\\/', basepath_re, 1), re.UNICODE)

        # make docs entrypage

        async def redirect_to_docs(data):
            response = RedirectResponse(url=self.config.basepath+'/docs')
            return response
        if self.config.basepath != "":
            self.app.routes.append(starlette_Route(
                self.config.basepath, redirect_to_docs))
        self.app.routes.append(starlette_Route(
            self.config.basepath+"/", redirect_to_docs))
        self.app.routes.append(starlette_Route(
            self.config.basepath+"/info", redirect_to_docs))                        

        # create model list endpoint
        @self.app.get(
            self.config.basepath+'/models',
            tags=["Metadata"],
            description="Get a list of all available models.",
            response_model=List[str]
        )
        async def models():
            return [m for m in self.model_dict]

        # get more information about a model
        @self.app.get(
            self.config.basepath+'/modelinfo/{model}',
            tags=["Metadata"],
            description="Get basic information of a model.",
            response_model=Dict[str, Any]
        )
        async def modelinfo(model: str):
            if model in self.model_dict:
                return self.model_dict[model].info()
            return {}

        # create an error dict endpont
        @self.app.get(
            self.config.basepath+'/errors',
            tags=["Metadata"],
            description="Get a dict of all errors.",
            response_model=Dict[str, Any]
        )
        async def errors():
            return self.error_dict

        # trigger a model refresh
        @self.app.get(
            self.config.basepath+'/refresh',
            tags=["Metadata"],
            description="Trigger the server to update the models."
        )
        async def refresh():
            self.init_scheduler()
            return "OK. Please be patient :)"

        #dummy route for the  eureka health call
        if self.config.eureka_server:
            @self.app.get(
                self.config.basepath+'/health',
                tags=["Metadata"],
                description="Eureka health call."
            )
            async def health():
                return {}
            @self.app.post(
                self.config.basepath+'/health',
                tags=["Metadata"],
                description="Eureka health call."
            )
            async def health():
                return {}

        # create model artifact endpoints

        # check caching
        if self.config.cache:
            self.full_cache_dir = os.path.join(
                pathlib.Path().absolute(), self.config.cachedir)
            logger.info(f"Use cachedir: {self.full_cache_dir}")
            pathlib.Path(self.full_cache_dir).mkdir(
                parents=True, exist_ok=True)
            self.update_models_from_cache()

    def check_token(self, token):
        """
        Check the request token
        """
        if len(self.config.token) == 0:
            return True
        if token.credentials in self.config.token:
            return True
        raise HTTPException(
            status_code=401,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )

    def get_version(self, m):
        prod_models = [
            mm for mm in m.latest_versions if mm.current_stage == "Production"]
        stage_models = [
            mm for mm in m.latest_versions if mm.current_stage == "Staging"]

        if self.config.staging and len(stage_models) > 0:
            return stage_models[0]
        if len(prod_models) > 0:
            return prod_models[0]
        return m.latest_versions[0]

    def update_models_from_cache(self):
        logger.info(f"Update models from cache")
        # load all available cached models
        for filename in os.listdir(self.full_cache_dir):
            newmodel = load_BaseHandler(os.path.join(
                self.full_cache_dir, filename), self)
            self.model_dict[newmodel.name] = newmodel

        self.app.openapi_schema = None

    def update_models(self):
        logger.info(f"Update models")

        all_models = []
        try:
            all_models = self.client.list_registered_models()
            if "server" in self.error_dict:
                del self.error_dict["server"]
        except Exception as ex:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            self.error_dict["server"] = {
                "message": str(ex),
                "type": str(exc_type),
                "file": fname,
                "line": exc_tb.tb_lineno,
                "exception": exc_obj
            }

        for m in all_models:

            # get model information
            name = urllib.parse.quote_plus(m.name)

            # get the best version
            model_version = self.get_version(m)

            # if the currently loaded model is already ok
            if name in self.model_dict and model_version.run_id == self.model_dict[name].run_id:
                continue

            # check if the type tag fits
            if len(self.config.tags) > 0 and all([not tt in m.tags.keys() for tt in self.config.tags]):
                continue

            logger.info(f"Update model {name}")

            # delete old route
            for idx2 in [idx for idx, r in enumerate(self.app.routes) if r.path == self.config.basepath+"/"+name]:
                del self.app.routes[idx2]

            # create new handler
            try:
                newmodel = BaseHandler(self, m, model_version)
                self.model_dict[name] = newmodel

                if name in self.error_dict:
                    del self.error_dict[name]

                if self.config.cache:
                    # create a cache representation of the new model
                    newmodel.save(os.path.join(self.full_cache_dir, name))

            except Exception as ex:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                self.error_dict[name] = {
                    "message": str(ex),
                    "type": str(exc_type),
                    "file": fname,
                    "line": exc_tb.tb_lineno,
                    "exception": exc_obj,
                    "traceback": traceback.format_exc().split("\n"),
                }

        self.app.openapi_schema = None

    def load_artifact(self, run_id, artifact_path):
        """ load an artifact from server, used for input_example
        """
        params = {"path": artifact_path, "run_uuid": run_id}
        res = requests.get(self.config.mlflow+"/get-artifact", params=params)
        return res

    def init_scheduler(self):
        """ init and restart the scheduler
        """
        if self.scheduler:
            try:
                self.scheduler.shutdown()
            except:
                pass

        self.scheduler = BackgroundScheduler()
        self.scheduler.add_job(func=self.update_models,
                               trigger="interval", seconds=self.config.timer)
        self.scheduler.add_job(func=self.update_models,
                               trigger="date",
                               run_date=datetime.datetime.now()
                               + datetime.timedelta(seconds=2))
        self.scheduler.start()



import os
import mlflow
from mlflow.types.schema import Schema
from pydantic import BaseModel
import numpy as np
import pandas as pd
from fastapi import HTTPException, Depends
from datetime import datetime

import dill as pickle
import importlib

import shutil
import pathlib



class BaseHandler:

    dtype_sample = {
        "float64": 1.234,
        "float32": 1.234,
        "int": 1,
        "int64": 1,
        "int32": 1,
        "str": "A",
        "object": "?"
    }

    def __init__(self, server,  m, model_version):
        self.server = server
        self.m = m
        self.name = m.name

        self.model_version_source = (model_version.source)
        model = mlflow.pyfunc.load_model(self.model_version_source)

        try:
            input_schema = model.metadata.get_input_schema()
        except:
            input_schema = None

        try:
            output_schema = model.metadata.get_output_schema()
        except:
            output_schema = None

        if not input_schema:
            input_schema = Schema([])
        if not output_schema:
            output_schema = Schema([])

        try:
            res = self.server.load_artifact(
                model_version.run_id,
                os.path.join(
                    model.metadata.artifact_path,
                    model.metadata.saved_input_example_info[
                        'artifact_path']
                ))
            print(res.json())
            self.input_example_data = res.json()['inputs']
        except:
            self.input_example_data = {}

        self.version = model_version.version
        self.source = model_version.source
        self.run_id = model_version.run_id
        self.model = model
        self.input_schema = input_schema if input_schema else {"inputs": []}
        self.output_schema = output_schema if output_schema else {"inputs": []}

        self.description = m.description
        timestamp = model_version.creation_timestamp/1000
        self.creation = datetime.fromtimestamp(
            timestamp).strftime('%Y-%m-%d %H:%M')

        self.long_description = f"""{m.description}\n\n"""
        try:
            if len(input_schema.inputs) > 0:
                self.long_description += f"<b>Input Schema:</b> {self.get_schema_string(input_schema)} <br/>\n"
        except:
            pass
        try:
            if len(output_schema.inputs) > 0:
                self.long_description += f"<b>Output Schema:</b> {self.get_schema_string(output_schema)}<br/>\n"
        except:
            pass

        self.long_description += f"""
<b>Version: </b> {self.get_version_link(m.name, model_version)}<br/>
<b>Run: </b> {self.get_experiment_link(m.name, model_version)}<br/>
<b>Creation: </b> {self.creation}
        """

        self.update_schema_classes()
        self.register_route()

    def update_schema_classes(self):

        if self.input_schema:
            input_schema_class = type(
                self.name+"-input",
                (BaseModel, ),
                {el["name"]:
                 self.input_example_data[el["name"]]
                 if el["name"] in self.input_example_data else
                 self.get_example_el(el) for el in self.input_schema.to_dict() if "name" in el})
        else:
            input_schema_class = None

        try:
            np_input = self.numpy_input(
                self.input_example_data, self.input_schema)
            output_example_data = self.model.predict(np_input)
            output_example_data = {k: v.tolist()
                                   for k, v in output_example_data.items()}
        except:
            output_example_data = {}

        if self.output_schema:
            output_schema_class = type(
                self.name+"-output",
                (BaseModel, ),
                {el["name"]:
                 output_example_data[el["name"]]
                 if el["name"] in output_example_data else
                 self.get_example_el(el) for el in self.output_schema.to_dict()})
        else:
            output_schema_class = None

        self.input_schema_class = input_schema_class() if input_schema_class else None
        self.output_schema_class = output_schema_class() if output_schema_class else None
        self.input_schema_class_type = input_schema_class
        self.output_schema_class_type = output_schema_class

    def register_route(self):
        if self.input_schema_class is None or len(self.input_schema.inputs) == 0:
            # no input create get interface
            if len(self.server.config.token) > 0:
                @self.server.app.get(
                    self.server.config.basepath+'/'+self.name,
                    description=self.long_description,
                    name=self.name, tags=["Models"],
                    response_model=self.output_schema_class_type
                )
                async def func(token: str = Depends(self.server.security)):
                    self.server.check_token(token)
                    return self.apply_model(None)
            else:
                @self.server.app.get(
                    self.server.config.basepath+'/'+self.name,
                    description=self.long_description,
                    name=self.name, tags=["Models"],
                    response_model=self.output_schema_class_type
                )
                async def func():
                    return self.apply_model(None)
        else:
            # create post interface
            if len(self.server.config.token) > 0:
                @self.server.app.post(
                    self.server.config.basepath+'/'+self.name,
                    description=self.long_description,
                    name=self.name, tags=["Models"],
                    response_model=self.output_schema_class_type
                )
                async def func(
                    data: self.input_schema_class_type,
                    token: str = Depends(self.server.security)
                ):
                    self.server.check_token(token)
                    return self.apply_model(data)
            else:
                @self.server.app.post(
                    self.server.config.basepath+'/'+self.name,
                    description=self.long_description,
                    name=self.name, tags=["Models"],
                    response_model=self.output_schema_class_type
                )
                async def func(data: self.input_schema_class_type):
                    return self.apply_model(data)

    def apply_model(self, data):
        # create a numpy input array
        if self.input_schema_class is None or len(self.input_schema.inputs) == 0:
            np_input = np.array([])
        else:
            try:
                np_input = self.numpy_input(data.__dict__, self.input_schema)
            except Exception as ex:
                raise self.get_error_message("Parse input error", ex)

        try:
            model_output = self.model.predict(np_input)
        except Exception as ex:
            raise self.get_error_message("Model prediction error", ex)

        try:
            output = self.parse_output(model_output)
        except Exception as ex:
            raise self.get_error_message("Parse output error", ex)

        return output

    def get_version_link(self, name, model_version):
        return f"{model_version.version}"

    def get_experiment_link(self, name, model_version):
        return f"{model_version.run_id}"

    def get_nested(self, dtype, shape):
        if len(shape) == 1:
            return [self.dtype_sample[dtype]]*shape[0]
        else:
            return [self.get_nested(dtype, shape[1:])]*max(1, shape[0])

    def get_example_el(self, el):
        if el["type"] == 'tensor':
            return self.get_nested(**el["tensor-spec"])
        return None

    def numpy_input(self, data, input_cfg):
        types = {el["name"]: el["tensor-spec"]["dtype"]
                 for el in input_cfg.to_dict() if "name" in el}
        return {
            key: np.array(val).astype(types[key])
            for key, val in data.items()
        }

    def get_error_message(self, loc, ex):
        self.server.logger.error(ex)
        return HTTPException(status_code=442, detail=[
            {
                "loc": [loc],
                "msg": str(ex),
                "type": str(type(ex))
            }
        ])

    def parse_output(self, data):
        if isinstance(data, pd.DataFrame):
            return data.to_dict(orient="list")
        return {
            key: val.tolist() if isinstance(val, (np.ndarray, np.generic)) else val
            for key, val in data.items()
        }

    def get_schema_string(self, schema):
        return "<ul><li>" + \
            '</li><li>'.join([
                '<b>'+s.name+'</b>: ' +
                str(s).replace('\''+s.name+'\':', '')
                for s in schema.inputs]) + \
            '</li></ul>'

    def info(self):
        return {
            "name": self.name,
            "version": self.version,
            "latest_versions": self.m.latest_versions,
            "input": self.input_schema.to_dict(),
            "output": self.output_schema.to_dict(),
            "description": self.description,
            "creation": self.creation
        }

    def save(self, dirname):
        """
        store the model to a directory
        """
        # save special references
        server = self.server
        model = self.model
        input_schema_class = self.input_schema_class
        output_schema_class = self.output_schema_class
        input_schema_class_type = self.input_schema_class_type
        output_schema_class_type = self.output_schema_class_type

        # null class references
        self.server = None
        self.model = None
        self.input_schema_class = None
        self.output_schema_class = None
        self.input_schema_class_type = None
        self.output_schema_class_type = None

        # create clean folder
        if os.path.exists(dirname):
            shutil.rmtree(dirname)
        pathlib.Path(dirname).mkdir(parents=True, exist_ok=True)

        # copy the mlflow model to a local folder
        extra_model_dir = os.path.join(dirname, "mlflow")
        pathlib.Path(extra_model_dir).mkdir(parents=True, exist_ok=True)
        self.extra_model_dir =_pyfunc_save_model_to_cache(self.model_version_source, extra_model_dir)

        # save the metadata
        filename = os.path.join(dirname, "meta.pkl")
        with open(filename, "wb") as f:
            pickle.dump(self, f)

        # reset class references
        self.server = server
        self.model = model
        self.input_schema_class = input_schema_class
        self.output_schema_class = output_schema_class
        self.input_schema_class_type = input_schema_class_type
        self.output_schema_class_type = output_schema_class_type


def load(dirname, server):
    """
    create a basehandler from cache
    """
    filename = os.path.join(dirname, "meta.pkl")
    with open(filename, "rb") as f:
        output = pickle.load(f)

    output.model = _pyfunc_load_model_from_cache(output.extra_model_dir)
    output.server = server
    output.update_schema_classes()
    output.register_route()

    return output


def _pyfunc_save_model_to_cache(model_uri, output_path):
    """
    download the artifact into the cache folder
    """

    local_path = mlflow.pyfunc._download_artifact_from_uri(artifact_uri=model_uri,
                                                           output_path=output_path)
    return local_path


def _pyfunc_load_model_from_cache(local_path):
    """
    load the pyfunc model directly from cache
    """
    from mlflow.models import Model
    from mlflow.pyfunc import PyFuncModel
    from mlflow.models.model import MLMODEL_FILE_NAME
    from mlflow.exceptions import MlflowException
    from mlflow.protos.databricks_pb2 import RESOURCE_DOES_NOT_EXIST
    FLAVOR_NAME = "python_function"
    DATA = "data"
    MAIN = "loader_module"
    CODE = "code"

    model_meta = Model.load(os.path.join(local_path, MLMODEL_FILE_NAME))

    conf = model_meta.flavors.get(FLAVOR_NAME)
    if conf is None:
        raise MlflowException(
            'Model does not have the "{flavor_name}" flavor'.format(
                flavor_name=FLAVOR_NAME),
            RESOURCE_DOES_NOT_EXIST,
        )

    data_path = os.path.join(local_path, conf[DATA]) if (
        DATA in conf) else local_path
    if CODE in conf and conf[CODE]:
        code_path = os.path.join(local_path, conf[CODE])
        mlflow.pyfunc.utils._add_code_to_system_path(code_path=code_path)        
    model_impl = importlib.import_module(conf[MAIN])._load_pyfunc(data_path)
    return PyFuncModel(model_meta=model_meta, model_impl=model_impl)

    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import mlflow_pyfunc_server
import uvicorn


def cli():
    config = mlflow_pyfunc_server.config.parse_known_args()[0]

    uvicorn.run("mlflow_pyfunc_server.serverapp:app",
                port=config.port,
                host=config.host,
                workers=config.workers
                )


if __name__ == '__main__':
    cli()



import os
from setuptools import setup, find_packages
import mlflow_pyfunc_server


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='mlflow_pyfunc_server',
    url="https://github.com/VK/mlflow-pyfunc-server",
    version=mlflow_pyfunc_server.__version__,
    packages=find_packages(),
    license = "MIT",
    keywords = "mlflow pyfunc",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    install_requires=[
        'mlflow',
        'uvicorn',
        'fastapi',
        'configargparse',
        'apscheduler'
    ],
    entry_points="""
        [console_scripts]
        mlflow-pyfunc-server=mlflow_pyfunc_server.cli:cli
    """
)
